from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.contenttypes.models import ContentType
from django.http.response import HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.text import slugify
from django.utils.translation import get_language
from django.views.generic import View

from cms.models import Page

from .constants import FILE_FORMATS, PAGE_EXPORTS


class PageExportView(UserPassesTestMixin, View):
    response_class = HttpResponse
    export_classes = PAGE_EXPORTS

    def test_func(self):
        return self.request.user.is_staff

    def get(self, request, *args, **kwargs):
        self.object = self.get_object(*args, **kwargs)
        self.language = get_language()
        self.file_format = FILE_FORMATS[kwargs.pop('file_format')]
        return self.render_to_response(request)

    def get_object(self, *args, **kwargs):
        return get_object_or_404(Page, pk=kwargs.pop('page_pk'))

    def render_to_response(self, request):
        export_class = self.export_classes[self.file_format.name]
        export_file = export_class(request, self.object, language=self.language).export()
        content_type = self.file_format.content_type

        response = self.response_class(export_file, content_type=content_type)
        response['Content-Disposition'] = 'attachment; filename="{}"'.format(
            self.get_file_name()
        )
        return response

    def get_file_name(self):
        title = self.object.get_title(language=self.language)

        return '{name}_{lang}.{ext}'.format(
            name=slugify(title),
            lang=self.language,
            ext=self.file_format.ext,
        )


class ModelExportView(PageExportView):

    def get_object(self, *args, **kwargs):
        app_name, model, pk = [kwargs.pop(arg) for arg in ('app', 'model', 'pk')]
        content_type = ContentType.objects.get_by_natural_key(app_name, model)
        return content_type.get_object_for_this_type(pk=pk)



from django.conf.urls import url

from .views import ModelExportView, PageExportView

app_name = 'export-page'

urlpatterns = [
    url(r'^(?P<page_pk>\d+)/(?P<file_format>\w+)/$', PageExportView.as_view(), name='cms_page'),
    url(r'^(?P<app>\w+):(?P<model>\w+)/(?P<pk>\d+)/(?P<file_format>\w+)/$', ModelExportView.as_view(), name='model'),
]



from collections import namedtuple

from django.contrib.sites.shortcuts import get_current_site
from django.db.models import CharField, ForeignKey, TextField

from cms.models import CMSPlugin, Placeholder, StaticPlaceholder
from cms.models.fields import PlaceholderField
from djangocms_page_meta.utils import get_page_meta

from djangocms_export_page import settings

from ..utils import clean_value

Section = namedtuple('Section', ['name', 'section', 'components'])
Component = namedtuple('Component', ['name', 'instance', 'fields'])
Field = namedtuple('Field', ['name', 'value'])

META_FIELD_BLACKLIST = ['_url']


class PageExport:
    """
    1. CMS Page don't need any extra configuration to work.

    2. If a Plugin has a ForeignKey that would behave like children,
    add the following to the CMSPlugin model class:

        _export_page = {
            'children': 'items'
        }

    where 'items' is a iterable attribute of the model class.
    And for on the ForeignKey Django model class:

        _export_page = {
            'fields': ['name', ... ]
        }

    3. If you need to export a Django model included in a AppHook,
    add the following to the model class:

        _export_page = {
            'sections': [{
                'name': 'Meta',
                'fields': ['title', ... ]
            }, {
                'name': 'Body',
                'fields': ['content']
            }],
        }

    It's better to put the PlaceholderField (here `content`) in a separate section.
    """

    FIELD_FILTERS = {
        'is_content': lambda f: isinstance(f, (CharField, TextField, PlaceholderField, ForeignKey)),
        'is_not_choice': lambda f: not f.choices,
        'is_not_slug': lambda f: f.name != 'slug',
        'is_not_link': lambda f: not f.name.endswith('_link')
    }

    def __init__(self, request, obj, language=None):
        self.request = request
        self.object = obj
        self.language = language
        self.base_url = self.get_base_url()
        self.page_url = self.get_page_url()
        self.page_meta = self.determine_meta_attributes(request, obj, language)

    def determine_meta_attributes(self, request, obj, language):
        page_meta = {}
        try: # If a Django CMS page
            meta_dict = get_page_meta(obj, language).__dict__
        except AttributeError:
            try: # If a Model object that uses the ModelMeta-mixin
                meta_dict = obj.get_meta(request)
            except Exception:
                meta_dict = {}

        # <meta property="some" content="value"> --> {'some (property)': value}
        page_meta.update({
            '{} ({})'.format(name, attr): value
            for attr, name, value in meta_dict.pop('extra_custom_props', [])
        })

        for key, value in meta_dict.items():
            if type(value) == list:
                page_meta[key] = ",".join(value)
            if type(value) == str and key not in META_FIELD_BLACKLIST:
                page_meta[key] = value
        return page_meta

    def get_base_url(self):
        return '{protocol}://{domain}'.format(
            protocol='https' if self.request.is_secure() else 'http',
            domain=get_current_site(self.request).domain
        )

    def get_page_url(self):
        return '{domain}{url}'.format(
            domain=self.get_base_url(),
            url=self.object.get_absolute_url()
        )

    def export(self):
        raise NotImplementedError

    def filter_fields(self, fields):
        def can_be_exported(field):
            return all((test(field) for test in self.FIELD_FILTERS.values()))

        return filter(can_be_exported, fields)

    def get_data(self):
        sections = self.get_sections()

        for name, section, components in sections:
            if isinstance(section, Placeholder):
                placeholder = section
                plugins = self.get_ordered_plugins(placeholder, self.language)
                for plugin in plugins:
                    components.extend(self.get_components(plugin))
            elif isinstance(section, StaticPlaceholder):
                placeholder = section
                plugins = self.get_ordered_plugins(placeholder.public, self.language)
                for plugin in plugins:
                    components.extend(self.get_components(plugin))

        return [s for s in sections if s.components]

    # 1. Section

    def get_section_name(self, placeholder=None):
        return placeholder.get_label()

    def get_sections(self):
        sections = []
        if hasattr(self.object, 'get_placeholders'):
            sections.extend(self.get_placeholders())
        if hasattr(self.object, '_export_page'):
            for section in self.object._export_page['sections']:
                sections.append(Section(
                    section['name'],
                    self.object,
                    self.get_defined_components(self.object, section['fields'])
                ))
        if hasattr(self.object, 'template'):
            codes = settings.EXPORT_STATIC_PLACEHOLDERS.get(self.object.template)
            if codes:
                queryset = StaticPlaceholder.objects.filter(code__in=codes)
                sections.extend(self.get_static_placeholders(queryset))

        if self.page_meta:
            sections.append(Section('Page meta-data', None, components=[
                Component('meta', None, fields=[
                    Field(key, value) for key, value in self.page_meta.items()
                ])]
            ))
        return sections

    def get_placeholders(self):
        placeholders = []

        for declared_placeholder in self.object.get_declared_placeholders(): # to always get the correct order
            placeholder = self.object.get_placeholders().get(slot=declared_placeholder.slot)
            name = self.get_section_name(placeholder)
            # Weird way to sort placeholders:
            if 'header' in placeholder.slot:
                placeholders.insert(0, Section(name, placeholder, []))
            else:
                placeholders.append(Section(name, placeholder, []))

        return placeholders

    def get_static_placeholders(self, static_placeholders):
        placeholders = []

        for placeholder in static_placeholders:
            placeholders.append(Section(placeholder.get_name(), placeholder, []))

        return placeholders

    @classmethod
    def get_ordered_plugins(cls, placeholder, language=None):
        yield from cls.next_plugin([
            CMSPlugin.objects.get(pk=pk)
            for pk in placeholder.get_plugin_tree_order(language)
        ])

    @classmethod
    def next_plugin(cls, plugins):
        for plugin in plugins:
            yield plugin
            yield from cls.next_plugin(plugin.get_children())

    # 2. Components

    def get_component_name(self, instance=None, plugin=None):
        name = plugin.name if plugin else instance._meta.verbose_name
        return '{}: {}'.format(name, instance)

    def get_components(self, obj):
        components = []

        try:
            instance, plugin = obj.get_plugin_instance()
        except AttributeError as error:
            # obj is a regular Django model instance
            instance, plugin = None, None
        if instance and plugin:
            components.append(Component(
                self.get_component_name(instance, plugin=plugin),
                instance,
                self.get_fields_with_values(instance, self.get_fields(instance))
            ))

        if hasattr(obj, '_export_page'):
            children_attr = obj._export_page.get('children', '')
            for child in getattr(obj, children_attr, []):
                components.extend(self.get_defined_components(child))
                components.extend(self.get_components(child))

        return components

    def get_defined_components(self, obj, field_names=None):
        """
        Gets components on a regular Django model object that are
        defined under `_export_page['fields']`.
        Usually returns one component, but can return one compontent for each Plugin
        in the placeholder fields.
        """
        field_names = field_names if field_names else obj._export_page['fields']
        model_fields = {f.name: f for f in obj._meta.fields}
        fields = [(name, model_fields[name]) for name in field_names]

        regular_fields = [f for f in fields if not isinstance(f[1], PlaceholderField)]
        placeholder_fields = [f for f in fields if isinstance(f[1], PlaceholderField)]

        components = [Component(
            self.get_component_name(obj),
            obj,
            self.get_fields_with_values(obj, regular_fields)
        )]

        for field_name, field in placeholder_fields:
            placeholder = getattr(obj, field_name)
            plugins = self.get_ordered_plugins(placeholder, self.language)
            for plugin in plugins:
                components.extend(self.get_components(plugin))

        return components

    # 3. Fields

    def get_fields_with_values(self, instance, fields):
        fields_with_values = []
        for field_name, field in fields:
            if field.many_to_one or field.one_to_one:
                relation = getattr(instance, field_name)
                if relation:
                    for relation_field in relation._meta.fields:
                        if hasattr(relation, '_export_page_field_names') and relation_field.name in relation._export_page_field_names:
                            value = getattr(relation, relation_field.name)
                            cleaned_data = clean_value(value)
                            if cleaned_data != None:
                                fields_with_values.append(Field(relation_field.verbose_name, cleaned_data))
            else:
                value = getattr(instance, field_name)
                cleaned_data = clean_value(value)
                if cleaned_data != None:
                    fields_with_values.append(Field(field.verbose_name, cleaned_data))
        return fields_with_values

    def get_fields(self, instance):
        fields = self.get_custom_fields(instance)
        return ((field.name, field) for field in fields)

    def get_custom_fields(self, instance):
        base_fields = self.get_base_fields(instance)
        fields = (f for f in instance._meta.fields if f.name not in base_fields)
        return self.filter_fields(fields)

    def get_base_fields(self, instance):
        is_regular_cms_plugin = hasattr(instance, 'cmsplugin_ptr')

        if is_regular_cms_plugin:
            fields = instance.cmsplugin_ptr._meta.fields
            yield 'cmsplugin_ptr'
        else:
            fields = instance._meta.fields

        for field in fields:
            yield field.name



from collections import namedtuple

from .export.docx import DocxPageExport

FileFormat = namedtuple('FileFormat', ['name', 'ext', 'content_type'])

DOCX = 'docx'
DOCX_CONTENT_TYPE = (
    'application/vnd'
    '.openxmlformats-officedocument.wordprocessingml.document'
)

FILE_FORMATS = {
    DOCX: FileFormat(DOCX, ext='docx', content_type=DOCX_CONTENT_TYPE),
}


PAGE_EXPORTS = {
    DOCX: DocxPageExport,
}



import unicodedata

from django.utils.html import strip_tags


def clean_value(value):
    """
    It would be better to use bleach.clean() as control chars as replaced
    by '? in bleach 2+.
    Requirement conflict: djangocms-text-ckeditor 3.6.0 needs
    html5lib<=0.9999999 but bleach needs the versions above that.
    Watch: https://github.com/divio/djangocms-text-ckeditor/issues/403
    """
    white_list = "\r\n"

    def is_control_char(char):
        return unicodedata.category(char).startswith("C")

    def valid_char(char):
        return not is_control_char(char) or char in white_list

    if isinstance(value, str):
        value = "".join(filter(valid_char, value))
        value = strip_tags(value).strip()
    return value



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from .constants import FILE_FORMATS


class ExportPageMixin:

    # def get_context_data(self, **kwargs):
    #     context = super().get_context_data(**kwargs)
    #     # This might be silly to call it there, but at least it’s late enough
    #     self.create_menu()
    #     return context

    # def create_menu(self):
    #     """
    #     Create a menu for your Model. Example code:

    #     menu = self.request.toolbar.get_or_create_menu(
    #         self.get_context_object_name(self.object),
    #         self.toolbar_menu_label
    #     )

    #     change_url = reverse_lazy(
    #         'admin:{m.app_label}_{m.model_name}_change'.format(m=self.model._meta),
    #         args=[self.object.id]
    #     )
    #     menu.add_sideframe_item(self.toolbar_change_label, url=change_url)

    #     self.create_export_page_menu_entry(menu)
    #     """

    #     try:
    #         super().create_menu()
    #     except AttributeError:
    #         raise NotImplementedError

    def create_export_page_menu_entry(self, current_page_menu):
        current_page_menu.add_break('export-page-break')
        file_formats = FILE_FORMATS.values()

        if len(file_formats) == 1:
            menu = current_page_menu
        else:
            menu = current_page_menu.get_or_create_menu('export-page', _('Export'))

        for file_format in file_formats:
            label = _('Export to {ext}').format(ext='.' + file_format.ext)
            export_url = reverse('export-page:model', kwargs={
                'app': self.object._meta.app_label,
                'model': self.object._meta.model_name,
                'pk': self.object.pk,
                'file_format': file_format.name})
            menu.add_link_item(label, url=export_url)



from collections import namedtuple
from io import BytesIO

from docx import Document
from docx.shared import Cm, Pt

from .common import PageExport

InlineStyle = namedtuple('InlineStyle', ['font', 'size', 'bold', 'italic'])


class DocxPageExport(PageExport):
    margin = Cm(2)
    font_name = 'Tahoma'
    table_style = 'TableGrid'
    column_layout = [Cm(2.5), Cm(7), Cm(8)]
    row_sytle = [InlineStyle(None, Pt(8), bold=True, italic=False), None, None]

    def setup_document(self):
        doc = Document()

        doc.styles['Normal'].font.name = self.font_name

        section = doc.sections[0]
        section.left_margin = section.right_margin = self.margin

        return doc

    def export(self):
        self.document = self.setup_document()
        self.document.add_heading(self.object.get_title(), level=2)
        self.document.add_paragraph().add_run(self.page_url).italic = True

        sections = self.get_data()

        for section_name, section, components in sections:
            self.document.add_heading(section_name, level=5)

            for component_name, instance, fields in components:
                rows = [self.get_row(field) for field in fields]

                if rows:
                    self.document.add_heading(component_name, level=8)
                    self.add_table(rows, has_header=False)

        return self.save()

    def save(self):
        handle = BytesIO()
        self.document.save(handle)
        handle.seek(0)
        return handle.read()

    def get_row(self, field):
        return [field.name, field.value, '']

    def add_table(self, rows, has_header=False, **kwargs):
        table = self.document.add_table(rows=0, cols=0)
        for width in self.column_layout:
            table.add_column(width)

        if has_header:
            self.populate_row(table, rows.pop(0))

        for row in rows:
            self.populate_row(table, row, **kwargs)

        table.style = self.table_style
        return table

    def populate_row(self, table, row, **kwargs):
        cells = table.add_row().cells

        for i, (value, style) in enumerate(zip(row, self.row_sytle)):
            self.apply_style(cells[i], str(value), style)

    def apply_style(self, obj, value, style):
        run = obj.paragraphs[0].add_run(value)
        if not style:
            return run
        if style.font:
            run.font = style.font
        if style.size:
            run.font.size = style.size
        if style.bold:
            run.bold = style.bold
        if style.italic:
            run.italic = style.italic
        return run



from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from cms.api import get_page_draft
from cms.toolbar.items import Break
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.utils.page_permissions import user_can_change_page

from .constants import FILE_FORMATS


@toolbar_pool.register
class ExportPageToolbar(CMSToolbar):

    def populate(self):
        page = get_page_draft(self.request.current_page)

        if not page or not user_can_change_page(self.request.user, page=page):
            return

        current_page_menu = self.toolbar.get_or_create_menu('page')
        position = self.get_position(current_page_menu)

        file_formats = FILE_FORMATS.values()

        if len(file_formats) == 1:
            menu = current_page_menu
        else:
            menu = current_page_menu.get_or_create_menu(
                'export-page', _('Export'), position=position)

        for file_format in file_formats:
            label = _('Export to {ext}').format(ext='.' + file_format.ext)
            url = reverse('export-page:cms_page', kwargs={
                'page_pk': page.pk,
                'file_format': file_format.name})
            menu.add_link_item(label, url=url, position=position)

        current_page_menu.add_break('export-page-break', position=position)

    def get_position(self, menu):
        # Last separator
        return menu.find_items(Break)[-1].index



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class DjangoCmsExportPageConfig(AppConfig):
    name = "djangocms_export_page"
    verbose_name = _("Django CMS Export Page")
